@@ -85,6 +85,9 @@ fi
 if ! which "$CC" ; then echo "Can't execute $CC" ; exit 1; fi
 if ! which "$CXX"; then echo "Can't execute $CXX"; exit 1; fi
 
+CFLAGS=${CFLAGS:-""}
+CXXFLAGS=${CXXFLAGS:-""}
+
 export CC CXX LD_LIBRARY_PATH
 
 EXTRA_SYSROOT=${EXTRA_SYSROOT:-""}
@@ -439,9 +442,11 @@ then
         BUILD_OPT+=" -DMYSQL_MAINTAINER_MODE=0"
         BUILD_OPT+=" -DWITH_ZLIB=system"
 
+        MYSQL_MM_VER="$MYSQL_MAJOR_VER$MYSQL_MINOR_VER"
+
         if [ "$MYSQL" == "mysql" ] # remove this distinction when MySQL
         then                       # fixes its SSL support
-            BUILD_OPT+="-DWITH_SSL=bundled"
+            BUILD_OPT+="-DWITH_SSL=yes"
         else
             BUILD_OPT+="-DWITH_SSL=system"
         fi
@@ -450,11 +455,10 @@ then
         then # MySQL-spcific build options
             BUILD_OPT+=" -DWITH_WSREP=1"
 
-            MYSQL_MM_VER="$MYSQL_MAJOR_VER$MYSQL_MINOR_VER"
-
-            [ "$MYSQL_MM_VER" -ge "56" ] && \
+            if [ "$MYSQL_MM_VER" -ge "56" ]
+            then
                 BUILD_OPT+=" -DWITH_LIBEVENT=yes -DWITH_INNODB_MEMCACHED=ON"
-
+            fi
             if [ "$MYSQL_MM_VER" -ge "57" ]
             then
                 BUILD_OPT+=" -DWITH_BOOST=boost_$MYSQL_MM_VER"
@@ -496,7 +500,8 @@ then
         cmake \
             -DCMAKE_C_COMPILER=$(basename $CC) \
             -DCMAKE_CXX_COMPILER=$(basename $CXX) \
-            -DCMAKE_CXX_FLAGS='-fpermissive' \
+            -DCMAKE_C_FLAGS="$CFLAGS" \
+            -DCMAKE_CXX_FLAGS="$CXXFLAGS" \
             -DBUILD_CONFIG=mysql_release \
             "${CMAKE_LAYOUT_OPTIONS[@]}" \
             $BUILD_OPT \
